/* 
 * Tcl_ErrorException.java --
 *
 *	This file defines the class Tcl_ErrorException which is used
 *      to signal Tcl error conditions within Java code.
 *
 * Copyright (c) 1996 by Sun Microsystems, Inc.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 * SCCS:
 */

package tcl;

/**
 * A Tcl result-code exception indicating an error condition.
 *
 * @version     1.0, 23 August 1996
 * @author      Glenn Vanderburg
 */

public
class Tcl_ErrorException extends Tcl_ResultCodeException 
{
    /**
     * A description of the error condition to appear in stack traces
     */
    private Value descrip;

    /**
     * Creates a new exception.
     * @param descrip a description of the error condition to appear in
     * stack traces
     */
    public
    Tcl_ErrorException(Value descrip) 
    {
	super("A Tcl error has occurred");
	this.descrip = descrip;
    }

    /**
     * Creates a new exception.
     * @param descrip a description of the error condition to appear in
     * stack traces
     */
    public
    Tcl_ErrorException(String descrip) 
    {
	super("A Tcl error has occurred");
	this.descrip = new Value(descrip);
    }

    /**
     * Creates a new exception.  This method is intended for use by
     * Tcl_UserException and its subclasses.
     * @param detail the standard detail string for exceptions
     * @param descrip a description of the error condition to appear in
     * stack traces
     */
    protected
    Tcl_ErrorException(String detail, Value descrip)
    {
	super(detail);
	this.descrip = descrip;
    }

    /**
     * The default constructor is not allowed.
     */
    private
    Tcl_ErrorException()
    {
	super("Something's wrong!");
    }

    /**
     * Returns the description of this error condition.
     * @return the Tcl description of this error condition.
     */
    public final Value
    getDescrip ()
    {
	return descrip;
    }

    /**
     * Returns the Tcl result code for this exception
     * @return the Tcl result code
     */
    public int
    getCode()
    {
	return Interp.TCL_ERROR;
    }
}
