/* 
 * TkApplication.java --
 *
 *	This file contains the TkApplication class.   This class
 *	is used to initialize an interpreter for use with Tk.
 *
 * Copyright (c) 1995 by Sun Microsystems, Inc.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 * SCCS: @(#) TkApplication.java 1.3 95/12/19 14:34:39
 */

package tcl;

/**
 * Holds useful methods for initializing and controlling a Tcl interpreter
 * for use with Tk.
 */

abstract public class TkApplication  {

    /**
     * Loads and initializes Tk into an interpreter, creating the main window.
     * @param interp the Tcl interpreter
     * @param name the Tk application name
     * @param classname the Tk application class name
     */
    public static native void tkInit(Interp interp,
	    String name, String classname);

    /**
     * Loads and initializes Tk into an interpreter, creating the main window.
     * Results in a call to tkInit(interp, "JTk", "JTk").
     * @param interp the Tcl interpreter
     */
    public static void tkInit(Interp interp) {
	tkInit(interp, "JTk", "JTk");
    }

    /**
     * Returns the number of main windows in this process.
     * @return the number of main windows in this process.
     */
    public static native int getNumMainWindows();

    /**
     * Runs the main event loop.
     * Loops processing events until all of the main windows for this
     * process have been destroyed.
     */
    public static void mainLoop() {
	while (getNumMainWindows() > 0) {
	    Interp.doOneEvent(0);
	}
    }
}
